/* -*-C-*-  
 *
 * parser.y --
 *
 *      This file contains the bison declarations that implement
 *      the 'expression' language for the Simple Database system.
 *      The syntax of the language is 'C'-like.  Expressions are
 *      used to define virtual fields, calculate new values into
 *      other database fields, and provide criteria to reselect
 *      and subset records.
 *
 *      The parser requires a lexical analyzer that breaks the
 *      input expressions up into tokens.  The analyzer is 
 *      specified in a 'flex' control file:  analyze.l.
 *
 * Copyright (c) 1995 Tom Moore.  All rights reserved.
 *
 * Permission is hereby granted, without written agreement and without
 * license or royalty fees, to use, copy, modify, and distribute this
 * software and its documentation for any purpose, provided that the
 * above copyright notice appear in all copies of this software.
 * 
 */
%{


#include <math.h>
#include "sdbInt.h"

/*
 * Definitions
 */
#define YYERROR_VERBOSE 1

/*
 * Linked lists to keep track of nodes allocated by
 * the parser.  Used for error recovery.
 */
typedef struct SdbNodeList {
  SdbNode            *node;
  struct SdbNodeList *next;
} SdbNodeList;

typedef struct SdbCatList {
  SdbCat             *cat;
  struct SdbCatList  *next;
} SdbCatList;

static SdbNodeList   *safeNodeList = NULL;
static SdbCatList    *safeCatList = NULL;

/*
 * Local variables used to pass the expression string to the
 * lexical analyzer.
 */
static int            sdbTextPos, sdbTextLen;
static char          *sdbTextBuff=NULL;

/*
 * Local variables that pass information into and out of the
 * parser.  Unfortunately, these are required because yyparse
 * does not allow arguments to be passed in.  This results in
 * a non-reentrant program!
 */
static Tcl_HashTable *htbl; 
static Tcl_Interp    *CurrentInterp=NULL;
static SdbNode       *expr;

/* Local prototypes */
static void           Sdberror _ANSI_ARGS_((char*));
int            Sdblex _ANSI_ARGS_((void));
static SdbNode       *safeGetNode _ANSI_ARGS_((void));
static SdbCat        *safeGetCat _ANSI_ARGS_((void));
static void           safeFreeMemory _ANSI_ARGS_((void));
static void           safeFreeLists  _ANSI_ARGS_((void));
int                   Sdbparse _ANSI_ARGS_((void));

#define YY_DECL static int     Sdblex _ANSI_ARGS_((void))


/*
 *------------------------------------------------------------
 *
 * Sdbwrap --
 *
 *      Called by the lexical analyzer at the end of input.
 *
 * Results:
 *      Tells the analyzer that there is no more input.
 *
 * Side effects:
 *      None.
 *
 *------------------------------------------------------------
 */
static int 
Sdbwrap _ANSI_ARGS_((void))
{
   return 1;
}


/*
 * The following memeory allocation routines are a pain in the
 * butt.  Unfortunately, they are required.  The bison/yacc parser
 * blows away the stack when it encounters an error, and that
 * is where the addresses to intermediate structures are stored.
 * The result is memory leaks when an error occurs.  My solution
 * is to keep track of the memory that is allocated.  If the
 * parse is successful, the list of allocations is erased.  If the parse
 * fails then the lists are traversed and the structures are freed.
 */

/*
 *------------------------------------------------------------
 *
 * safeGetNode --
 *
 *      Called by the parser to get a node and remember the
 *      location
 *
 * Results:
 *      Returns an empty node.
 *
 * Side effects:
 *      Adds the address of the node onto the node list.
 *
 *------------------------------------------------------------
 */
static SdbNode        *
safeGetNode _ANSI_ARGS_((void))
{
  SdbNode *node=SdbGetNode();
  SdbNodeList *memo=(SdbNodeList*)ckalloc(sizeof(SdbNodeList));
  memo->node=node;
  memo->next = safeNodeList;
  safeNodeList = memo;
  return(node);
}

static SdbCat         *
safeGetCat _ANSI_ARGS_((void))
{
  SdbCat *cat=(SdbCat*) ckalloc(sizeof(SdbCat));
  SdbCatList *memo=(SdbCatList*)ckalloc(sizeof(SdbCatList));
  memo->cat=cat;
  memo->next = safeCatList;
  safeCatList = memo;
  return(cat);
}


/*
 *------------------------------------------------------------
 *
 * safeFreeMemory --
 *
 *      Called by the parser free all nodes that have been
 *      allocated in this parse.
 *
 * Results:
 *      None
 *
 * Side effects:
 *      None
 *
 *------------------------------------------------------------
 */
static void
safeFreeMemory _ANSI_ARGS_((void))
{
  SdbNodeList *memo=safeNodeList;
  SdbCatList *catList=safeCatList;
  SdbNode *tree;
  
  while(memo) {
    tree=memo->node;
    switch (tree->type) {
    case hsgCodeList:
        ckfree(tree->content.list.pntr);
        break;

    case curveFnct:{
            if (tree->content.yieldHandle->list)
	        ckfree(tree->content.yieldHandle->list);
            ckfree(tree->content.yieldHandle);
            break;
        }

    default:
        break;
    }
    ckfree(tree);
    memo = memo->next;
  }
  while(catList) {
    ckfree(catList->cat);
    catList = catList->next;
  }
}

static void
safeFreeLists  _ANSI_ARGS_((void))
{  
  SdbNodeList *memo=safeNodeList, *temp;
  SdbCatList *catList=safeCatList, *temp2;
  while(memo) {
    temp=memo->next;
    ckfree(memo);
    memo = temp;
  }
  safeNodeList = NULL;
  while(catList) {
    temp2=catList->next;
    ckfree(catList);
    catList = temp2;
  }
  safeCatList = NULL;
}
    

%}

/*
 *------------------------------------------------------------
 *
 * Sdbparse --
 *
 *      This code is interpolated by 'bison' or 'yacc'
 *      into the Sdbparse routine.
 *
 * Results:
 *      Returns the parse tree.
 *
 * Side effects:
 *      None.
 *
 *------------------------------------------------------------
 */

/*
 *------------------- bison declarations --------------------
 */

%union {
    SdbNode      *node;
    int          integer;
    SdbOpIndex   op;
    double       value;
    double       (*fnctptr)();
    char         *string;
    SdbCat       *classList;
}


%token IF IN RESERVED FORMAT INTERPOLATE EOS EOI CONNECTOR RANDOM
%token <value>   FLOAT                              /* constant */
%token <fnctptr> C_FNCT S_FNCT  D_FNCT HSG_FNCT     /* intrinsic functions */
%token <string>  HSG_CODE_SYM HSG_NUM_SYM           /* string pointer */
%token <string>  UNKNOWN MULTI_STRING STRING
%token <integer> INT


%type <node>     expr allExpr list allStrings       /* function expression */
%type <node>     strings args oneArg multiStatement /* function expression */
%type <value>    constValue
%type <op>       logicalRelation booleanRelation 
%type <classList> numChrCat oneNumChrCat 
%type <classList> numNumCat oneNumNumCat 
%type <classList> chrChrCat oneChrChrCat 
%type <classList> chrNumCat oneChrNumCat 

%left '='
%left <op> OR 
%left <op> AND 
%nonassoc '?' 
%nonassoc numc
%left <op>  EQ NE 
%left <op> LT LE GT GE
%left  MINUS PLUS
%left  MUL DIV MOD
%left  SIG
%right POW
%right NOT
%right UNARY
%right IN

%start input

%% 

/*
 * ------------------- syntax rules ------------------------ 
 */ 

/*
 * The starting input rule.  
 */
input : 

/* Match an empty input line... */
EOI               {  return 0; }

/* ...or a line containing a string expression... */
| allStrings EOI {
expr = $1;
YYACCEPT;
}

/* ...or a line containing a numeric expression... */
| allExpr EOI {
expr = $1;
YYACCEPT;
}

/* ...or in general, trap all errors */
| error EOI {
  YYABORT;
}

| allStrings error EOI {
  YYABORT;
} 
| allExpr error EOI {
  YYABORT;
} 
;

/*
 * The empty rule.
 */
 empty: {;}
;

/*
 * A constant value
 */
 constValue: FLOAT { $$ = $1;}
      | INT   { $$ = (double) $1;}
;

/*
 * Rule for numeric expressions.  A numeric expression can
 * be expression or an error.
 */ 
 allExpr:
 expr %prec '=' {
  $$ = $1;
}

/* trap an error */
| HSG_NUM_SYM '=' allStrings {
Sdberror("cannot assign code value to numeric variable");
YYERROR;
}
;

/*
 * Rule for character expressions.  A character expression can
 * be an expression or an error.
 */ 
 allStrings:
strings %prec '=' {
  $$ = $1;
}

/* trap an error */
| HSG_CODE_SYM '=' expr {
Sdberror("cannot assign numeric value to code variable");
YYERROR;
}
;


/*
 *  Rules for character conditional expressions of the
 *  form   (expr)?(op,expr,string),...,string
 *  This rule is helped by the 'oneNumChrCat' rule.
 */
 numChrCat:  oneNumChrCat strings %prec '?' {
   SdbCat *pntr, *newclass=(SdbCat*) ckalloc(sizeof(SdbCat));

   /* create a new character node */
   if (!newclass) {
     fprintf(stderr,"Out of memory\n");
     exit(1);
   }

   /* add the default value to the end of the list of categories */
   for (pntr = $1; pntr->next; pntr=pntr->next);

   /* set the appropriate values */
   pntr->next = newclass;
   newclass->expr = NULL;
   newclass->relation = opNop;
   newclass->code = $2;
   newclass->next = NULL;

   /* return the pointer to the beginning of the structure */
   $$ = $1;
 };

/*
 * The oneNumChrCat rule recursively parses out the conditionals in the form
 *   ...,(operator,expr,string)
 */
 oneNumChrCat: '(' logicalRelation ',' expr ',' strings ')' CONNECTOR
 {
   SdbCat *newclass= safeGetCat();

   if (!newclass) {
     fprintf(stderr,"Out of memory\n");
     exit(1);
   }
   
   newclass->code = $6;
   newclass->relation = $2;
   newclass->expr = $4;
   newclass->next = NULL;
   $$ = newclass;
   
 }
|  oneNumChrCat '(' logicalRelation ',' expr ',' strings ')' CONNECTOR 
{    
   SdbCat *pntr, *newclass=(SdbCat*) ckalloc(sizeof(SdbCat));
 
   if (!newclass) {
     fprintf(stderr,"Out of memory\n");
     exit(1);
   }
   /* add this to the end of the previous list */
   for (pntr = $1; pntr->next; pntr=pntr->next);

   pntr->next=newclass;
   newclass->code = $7;
   newclass->relation = $3;
   newclass->expr = $5;
   newclass->next = NULL;
   $$ = $1;
   }
;

/*
 * This rule is similar to the previous, but it is a numeric
 * conditional expression, of the form:
 *    (expr)?(operator,expr,expr),...,expr
 * This rule is helped by the oneNumNumCat rule.
 */
 numNumCat:  oneNumNumCat expr %prec numc {
   SdbCat *pntr, *newclass=(SdbCat*)ckalloc(sizeof(SdbCat));

   /* create a new character node */
   if (!newclass) {
     fprintf(stderr,"Out of memory\n");
     exit(1);
   }

   /* add the default value to the end of the list of categories */
   for (pntr = $1; pntr->next; pntr=pntr->next);

   /* set the appropriate values */
   pntr->next = newclass;
   newclass->expr = NULL;
   newclass->relation = opNop;
   newclass->code = $2;
   newclass->next = NULL;

   /* return the pointer to the beginning of the structure */
   $$ = $1;
 };

 oneNumNumCat: '(' logicalRelation ',' expr ',' expr ')' CONNECTOR
 {
   SdbCat *newclass = safeGetCat();

   if (!newclass) {
     fprintf(stderr,"Out of memory\n");
     exit(1);
   }
   
   newclass->code = $6;
   newclass->relation = $2;
   newclass->expr = $4;
   newclass->next = NULL;
   $$ = newclass;
   
 }
|  oneNumNumCat '(' logicalRelation ',' expr ',' expr ')' CONNECTOR 
{    
   SdbCat *pntr,*newclass= (SdbCat*)ckalloc(sizeof(SdbCat));

   if (!newclass) {
     fprintf(stderr,"Out of memory\n");
     exit(1);
   }
   /* add this to the end of the previous list */
   for (pntr = $1; pntr->next; pntr=pntr->next);

   pntr->next=newclass;
   newclass->code = $7;
   newclass->relation = $3;
   newclass->expr = $5;
   newclass->next = NULL;
   $$ = $1;
   }
;

/*
 *  Rules for character conditional expressions of the
 *  form   (string)?(op,string,string),...,string
 *  This rule is helped by the 'oneChrChrCat' rule.
 */
 chrChrCat:  oneChrChrCat strings %prec '?' {
   SdbCat *pntr, *newclass=(SdbCat*) ckalloc(sizeof(SdbCat));

   /* create a new character node */
   if (!newclass) {
     fprintf(stderr,"Out of memory\n");
     exit(1);
   }

   /* add the default value to the end of the list of categories */
   for (pntr = $1; pntr->next; pntr=pntr->next);

   /* set the appropriate values */
   pntr->next = newclass;
   newclass->expr = NULL;
   newclass->relation = opNop;
   newclass->code = $2;
   newclass->next = NULL;

   /* return the pointer to the beginning of the structure */
   $$ = $1;
 };

/*
 * The oneChrChrCat rule recursively parses out the conditionals in the form
 *   ...,(operator,expr,string)
 */
 oneChrChrCat: '(' logicalRelation ',' strings ',' strings ')' CONNECTOR
 {
   SdbCat *newclass= safeGetCat();

   if (!newclass) {
     fprintf(stderr,"Out of memory\n");
     exit(1);
   }
   
   newclass->code = $6;
   newclass->relation = $2;
   newclass->expr = $4;
   newclass->next = NULL;
   $$ = newclass;
   
 }
|  oneChrChrCat '(' logicalRelation ',' strings ',' strings ')' CONNECTOR 
{    
   SdbCat *pntr, *newclass=(SdbCat*) ckalloc(sizeof(SdbCat));
 
   if (!newclass) {
     fprintf(stderr,"Out of memory\n");
     exit(1);
   }
   /* add this to the end of the previous list */
   for (pntr = $1; pntr->next; pntr=pntr->next);

   pntr->next=newclass;
   newclass->code = $7;
   newclass->relation = $3;
   newclass->expr = $5;
   newclass->next = NULL;
   $$ = $1;
   }
;

/*
 * This rule is similar to the previous, but it is a character-numeric
 * conditional expression, of the form:
 *    (string)?(operator,string,expr),...,expr
 * This rule is helped by the oneChrNumCat rule.
 */
 chrNumCat:  oneChrNumCat expr %prec numc {
   SdbCat *pntr, *newclass=(SdbCat*)ckalloc(sizeof(SdbCat));

   /* create a new character node */
   if (!newclass) {
     fprintf(stderr,"Out of memory\n");
     exit(1);
   }

   /* add the default value to the end of the list of categories */
   for (pntr = $1; pntr->next; pntr=pntr->next);

   /* set the appropriate values */
   pntr->next = newclass;
   newclass->expr = NULL;
   newclass->relation = opNop;
   newclass->code = $2;
   newclass->next = NULL;

   /* return the pointer to the beginning of the structure */
   $$ = $1;
 };

 oneChrNumCat: '(' logicalRelation ',' strings ',' expr ')' CONNECTOR
 {
   SdbCat *newclass = safeGetCat();

   if (!newclass) {
     fprintf(stderr,"Out of memory\n");
     exit(1);
   }
   
   newclass->code = $6;
   newclass->relation = $2;
   newclass->expr = $4;
   newclass->next = NULL;
   $$ = newclass;
   
 }
|  oneChrNumCat '(' logicalRelation ',' strings ',' expr ')' CONNECTOR 
{    
   SdbCat *pntr,*newclass= (SdbCat*)ckalloc(sizeof(SdbCat));

   if (!newclass) {
     fprintf(stderr,"Out of memory\n");
     exit(1);
   }
   /* add this to the end of the previous list */
   for (pntr = $1; pntr->next; pntr=pntr->next);

   pntr->next=newclass;
   newclass->code = $7;
   newclass->relation = $3;
   newclass->expr = $5;
   newclass->next = NULL;
   $$ = $1;
   }
;
 
/*
 * A rule to implement statement sequences.  Sequences are strings
 * of statements seperated by ';'.  The value of the sequence is
 * the value of the last statement in the sequence.   This is
 * only half of the statement sequence rule.  The other parts are
 * implemented in the 'expr' and 'strings' rules.
 */
 multiStatement:

/* A statement sequence can be just a numeric expression... */
allExpr EOS {
  SdbNode *newnode=safeGetNode();
  newnode->type = numSequence;
  newnode->left = NULL;
  newnode->right = $1;
  $$ = newnode;
}

/* ...or a character expression... */
| allStrings EOS {
  SdbNode *newnode=safeGetNode();
  newnode->type = charSequence;
  newnode->left = NULL;
  newnode->right = $1;
  $$ = newnode;
}

/* ...or a statement sequence followed by a numeric expression... */
| multiStatement allExpr EOS {
  SdbNode *newnode=safeGetNode();
  newnode->type = numSequence;
  newnode->left = $1;
  newnode->right = $2;
  $$ = newnode;
}

/* ...or a statement sequence followed by a character expression... */
| multiStatement allStrings EOS {
  SdbNode *newnode=safeGetNode();
  newnode->type = charSequence;
  newnode->left = $1;
  newnode->right = $2;
  $$ = newnode;
}
;

/*
 * Rules for numeric expressions.  There are lots of these!
 * Most follow the syntax of 'C'.
 */
expr  : 

/* assignment of an expression to a symbol */
  HSG_NUM_SYM '=' expr {
  SdbNode *subnode, *newnode=safeGetNode();
  newnode->type = assignNum;
  newnode->right = $3;
  subnode = safeGetNode();
  subnode->type = hsgNumber;
  subnode->content.code = $1;
  subnode->right=NULL;
  subnode->left=NULL;
  newnode->left=subnode;
  $$ = newnode;
}

/* Trap use of an unknown token name... */
| UNKNOWN {
  Sdberror("cannot use unknown symbol in expression");
  YYERROR;
}

/* ...the last part of a statement sequence... */
| '(' multiStatement allExpr ')' {
  SdbNode *newnode=safeGetNode();
  newnode->type = numSequence;
  newnode->left = $2;
  newnode->right = $3;
  $$ = newnode;
}

/* ...a constant value...*/
| constValue {
  SdbNode *newnode=safeGetNode();
  newnode->type = constValue;
  newnode->content.value = $1;
  $$ = newnode;
}

/* ...an existing numeric variable... */
| HSG_NUM_SYM {
  SdbNode *newnode=safeGetNode();
  newnode->type = hsgNumber;
  newnode->content.code = $1;
  $$ = newnode;
}

/* ...a specialized HSG function... */
| HSG_FNCT  '(' strings ')' {
  SdbNode *newnode=safeGetNode();
  newnode->type = hsgFnct;
  newnode->content.fnctptr = $1;
  newnode->right = $3;
  $$ = newnode;
}

/* ...a simplex math function (e.g. cos, sin) ... */
| S_FNCT '(' expr ')'  { 
  SdbNode *newnode=safeGetNode();
  newnode->type = simplexFnct;
  newnode->content.fnctptr = $1;
  newnode->right = $3;
  $$ = newnode;
}

/* ...a numeric function that takes a single character argument (e.g. atof) */
| C_FNCT '(' strings ')'  { 
  SdbNode *newnode=safeGetNode();
  newnode->type = charFnct;
  newnode->content.fnctptr = $1;
  newnode->right = $3;
  $$ = newnode;
}

/* ...a duplex math function (e.g. pow(x,y))... */
| D_FNCT '(' expr ',' expr ')' {
  SdbNode *newnode=safeGetNode();
  newnode->type = duplexFnct;
  newnode->content.fnctptr = $1;
  newnode->left = $3;
  newnode->right= $5;
  $$ = newnode;
}

/* ...interpolate along a yield curve... */
| INTERPOLATE '(' STRING ',' strings ',' strings ',' expr ')' {
  SdbNode *newnode=safeGetNode(), *subnode, *tempnode;
  newnode->type = curveFnct;
  newnode->content.yieldHandle = 
    (SdbYieldHandle*)ckalloc(sizeof(SdbYieldHandle));
  (newnode->content.yieldHandle)->xIndex = -1;
  (newnode->content.yieldHandle)->yIndex = -1;
  (newnode->content.yieldHandle)->yieldSet = NULL;
  (newnode->content.yieldHandle)->list     = NULL;
  (newnode->content.yieldHandle)->stamp    = 0;
  subnode=safeGetNode();
  tempnode= safeGetNode();
  tempnode->type = noScan;
  tempnode->left=safeGetNode();
  tempnode->left->type = constCode;
  tempnode->left->content.code = $3;
  tempnode->right = NULL;
  subnode->type = binaryOp;
  subnode->content.op =opNop;
  subnode->left=tempnode;
  subnode->right=$5;
  newnode->left = subnode;
  subnode=safeGetNode();
  subnode->type = binaryOp;
  subnode->content.op =opNop;
  subnode->left=$7;
  subnode->right=$9;
  newnode->right = subnode;
  $$ = newnode;
}

| INTERPOLATE '(' STRING  allStrings ',' strings ',' strings ',' expr ')' {
  SdbNode *newnode, *subnode, *tempnode;
  
  /*
   * Before accepting this node, make sure that the $4 expr is a
   * character sequence consisting only of simple variable assignments.
   */
  tempnode=$4;
  if (tempnode->type == assignChar) {
    if (tempnode->right->type == hsgCode) 
      /* ok, do nothing */;
    else {
      Sdberror("lookup alias must be simple variable assignment");
      YYERROR;
    }
  }
  else if (tempnode->type == charSequence) {
    SdbNode *next=tempnode;
    while (next) {
      if (!(next->right->type == assignChar &&
	    next->right->right->type == hsgCode)) {
	Sdberror("lookup alias must be sequence of simple variable assignments");
	YYERROR;
      }
      next = next->left;
    }
  }
  else {
    Sdberror("lookup alias must be sequence of simple variable assignments");
    YYERROR;
  }

  /*
   * The alias expression looks ok.  Process the rest of the 
   * record.
   */

  newnode=safeGetNode();
  newnode->type = curveFnct;
  newnode->content.yieldHandle = 
    (SdbYieldHandle*)ckalloc(sizeof(SdbYieldHandle));
  (newnode->content.yieldHandle)->xIndex = -1;
  (newnode->content.yieldHandle)->yIndex = -1;
  (newnode->content.yieldHandle)->yieldSet = NULL;
  (newnode->content.yieldHandle)->list     = NULL;
  (newnode->content.yieldHandle)->stamp    = 0;
  subnode=safeGetNode();
  tempnode= safeGetNode();
  tempnode->type = noScan;
  tempnode->left=safeGetNode();
  tempnode->left->type = constCode;
  tempnode->left->content.code = $3;
  tempnode->right = $4;
  subnode->type = binaryOp;
  subnode->content.op =opNop;
  subnode->left=tempnode;
  subnode->right=$6;
  newnode->left = subnode;
  subnode=safeGetNode();
  subnode->type = binaryOp;
  subnode->content.op =opNop;
  subnode->left=$8;
  subnode->right=$10;
  newnode->right = subnode;
  $$ = newnode;
}

/* ... a random number stream ... */
| RANDOM '(' STRING ')' {
  SdbNode *subnode=safeGetNode(), *newnode=safeGetNode();
  newnode->type = randomFnct;
  newnode->content.randomPtr = NULL;
  subnode->type = constCode;
  subnode->content.code = $3;
  newnode->right = subnode;
  newnode->left=NULL;
  $$ = newnode;
}

/* ... a list find routine ... */
| IN '(' strings ',' list ')' %prec LE {
  SdbNode *newnode=safeGetNode();
  newnode->type = stringFind;
  newnode->left = $3;
  newnode->right= $5;
  $$ = newnode;
}

/* ... expression comparisons ... */
| expr logicalRelation expr %prec LE     {
  SdbNode *newnode=safeGetNode();
  newnode->type = binaryOp;
  newnode->content.op = $2;
  newnode->left = $1;
  newnode->right= $3;
  $$ = newnode;
}

/* ... boolean relations ... */
| expr booleanRelation expr %prec AND  {
  SdbNode *newnode=safeGetNode();
  newnode->type = booleanOp;
  newnode->content.op = $2;
  newnode->left = $1;
  newnode->right= $3;
  $$ = newnode;
}

/* ... addition ... */
| expr PLUS expr      {
  SdbNode *newnode=safeGetNode();
  newnode->type = binaryOp;
  newnode->content.op = opSum;
  newnode->left = $1;
  newnode->right= $3;
  $$ = newnode;
}

/* ... subtraction ... */
| expr MINUS expr      {
  SdbNode *newnode=safeGetNode();
  newnode->type = binaryOp;
  newnode->content.op = opSub;
  newnode->left = $1;
  newnode->right= $3;
  $$ = newnode;
}

/* ... multiplication ... */
| expr MUL expr      {
  SdbNode *newnode=safeGetNode();
  newnode->type = binaryOp;
  newnode->content.op = opMul;
  newnode->left = $1;
  newnode->right= $3;
  $$ = newnode;
}

/* ... division ... */
| expr DIV expr      {
  SdbNode *newnode=safeGetNode();
  newnode->type = binaryOp;
  newnode->content.op = opDiv;
  newnode->left =$1;
  newnode->right=$3;
  $$ = newnode;
}

/* ... exponetiation ... */
| expr POW expr  {
  SdbNode *newnode=safeGetNode();
  newnode->type = duplexFnct;
  newnode->content.fnctptr = pow;
  newnode->left = $1;
  newnode->right= $3;
  $$ = newnode;
}

/* ... modulus ... */
| expr MOD expr      {
  SdbNode *newnode=safeGetNode();
  newnode->type = duplexFnct;
  newnode->content.fnctptr = fmod;
  newnode->left = $1;
  newnode->right= $3;
  $$ = newnode;
}

/* ... string comparisons (result in 0 or 1) ... */
| strings logicalRelation strings %prec LE {
  SdbNode *newnode=safeGetNode();
  newnode->type = stringCompare;
  newnode->content.op = $2;
  newnode->left = $1;
  newnode->right= $3;
  $$ = newnode;
}

/* ... unary minus (negation) ... */
| MINUS expr %prec UNARY {
  SdbNode *newnode=safeGetNode();
  newnode->type = unaryOp; 
  newnode->content.op = opNeg;
  newnode->right = $2;
  $$ = newnode;
}

/* ...unary plus (no operation) ... */
| PLUS expr %prec UNARY {
 $$ = $2;
}

/* ... logical not ... */
| NOT expr %prec UNARY {
  SdbNode *newnode=safeGetNode();
  newnode->type = unaryOp;
  newnode->content.op = opNot;
  newnode->right= $2;
  $$ = newnode;
}

/* ... numeric conditional expression returning numeric ... */
| expr '?' numNumCat   %prec LE        {  
  SdbNode *newnode=safeGetNode();
  newnode->type = numNumFnct;
  newnode->left = $1;
  newnode->content.classList = $3;
  $$ = newnode;
} 

/* ... character conditional expression returning numeric ... */
| strings '?' chrNumCat   %prec LE        {  
  SdbNode *newnode=safeGetNode();
  newnode->type = chrNumFnct;
  newnode->left = $1;
  newnode->content.classList = $3;
  $$ = newnode;
} 

/* ... parenthetical grouping ... */
| '(' expr ')'  %prec IN { 
  $$ = $2; 
}
;

/*
 * Rule for lists. Split the string up into smaller chunks.
 * Currently this is only used in the string find function.
 * I may change this in order to make it compatible with 
 * Tcl lists.
 */
list:  MULTI_STRING {
   SdbNode *newnode = safeGetNode();
   char **pntr, *code, *token = $1;
   int k=1, new;

   /* make space for a fixed length code list */
   for (code=token; code[0] != 0; code++) 
     if (code[0] == '/') k++;

   if (!(pntr = (char**)ckalloc(k*sizeof(char*)))) {
     fprintf(stderr,"Out of memory\n");
     exit(1);
   }
   
   /* loop through and load up the list */
   k = 0;
   while (token && ((code = strtok (token, "/")) != NULL)) {
     pntr[k] = GetIndices (htbl,code, &new);
     token = strtok (NULL, "");
     k++;
   }
   newnode->type = hsgCodeList;
   newnode->content.list.n = k;
   newnode->content.list.pntr = pntr;
   ckfree($1);
   $$ = newnode;
 }

 /* functions returning lists should also be defined here */
;

/*
 * Rules for anything that results in a string value. 
 */
 strings: 

/* A string can be an assignemt of a string... */
  HSG_CODE_SYM '=' strings {
  SdbNode *subnode, *newnode=safeGetNode();
  newnode->type = assignChar;
  newnode->right = $3;
  subnode = safeGetNode();
  subnode->type = hsgCode;
  subnode->content.code = $1;
  subnode->right=NULL;
  subnode->left=NULL;
  newnode->left=subnode;
  $$ = newnode;
}

/* ... or a statement sequence ending in a string expression... */
|'(' multiStatement allStrings ')' {
  SdbNode *newnode=safeGetNode();
  newnode->type = charSequence;
  newnode->left = $2;
  newnode->right = $3;
  $$ = newnode;
}

/* ... or a string constant ... */
| STRING {
  SdbNode *newnode=safeGetNode();
  newnode->type = constCode;
  newnode->content.code = $1;
  $$ = newnode;
} 

/* ... or a already existing character variable ... */
| HSG_CODE_SYM {
  SdbNode *newnode=safeGetNode();
  newnode->type = hsgCode;
  newnode->content.code = $1;
  $$ = newnode;
}

/* ... or a numeric conditional expression returning a character ... */
| expr '?' numChrCat   %prec LE        {  
  SdbNode *newnode=safeGetNode();
  newnode->type = numChrFnct;
  newnode->left = $1;
  newnode->content.classList = $3;
  $$ = newnode;
} 

/* ... or a character conditional expression returning a character ... */
| expr '?' chrChrCat   %prec LE        {  
  SdbNode *newnode=safeGetNode();
  newnode->type = chrChrFnct;
  newnode->left = $1;
  newnode->content.classList = $3;
  $$ = newnode;
} 

/* ... or a format function (similar to sprintf) ... */
| FORMAT '(' strings args ')' {
/* This function takes a fixed string as input.  It could be a variable
   string, but that would complicate things (especially if it were another
   format function, because the format function is not re-entrant) */
  SdbNode *newnode=safeGetNode();
  newnode->type = formatFnct;
  newnode->left = $3;
  newnode->right = $4;
  $$ = newnode;
}

/* ... parenthetical grouping ... */
| '(' strings ')'  %prec IN { 
  $$ = $2; 
}
;


/*
 * Miscellaneous rules.  Little bits and pieces needed to make the
 * rest fits together.
 */


/* 
 * The next two are helper functions for the format function, to parse
 * a variable number of arguments.
 */
 args: args ',' oneArg {

   if ($1) {
     $1->right = $3;
     $$ = $1;
   }
   else
     $$ = $3;
   
 }

| empty { $$ = NULL; } 
;

oneArg : expr {
  SdbNode *newnode=safeGetNode();
  newnode->type = numExpr;
  newnode->left = $1;
  newnode->right = NULL;
  $$ = newnode;

}
| strings {  
  SdbNode *newnode=safeGetNode();
  newnode->type = charExpr;
  newnode->left = $1;
  newnode->right = NULL;
  $$ = newnode;
}

;

/*
 * Logical relations
 */
 logicalRelation: 
  EQ %prec EQ { $$ = opEq; }
| LT %prec LT { $$ = opLt; }
| GT %prec GT { $$ = opGt; }
| GE %prec GE { $$ = opGe; }
| LE %prec LE { $$ = opLe; }
| NE %prec NE { $$ = opNe; }
;

 booleanRelation: 
  AND %prec AND { $$ = opAnd; }
| OR  %prec OR  { $$ = opOr; } 
;

/* Error expressions */

expr:  ')' %prec '=' {
   Sdberror("unmatched right parenthesis");
   YYERROR;
 }

|  '(' EOI %prec '=' {
  Sdberror("unmatched left parenthesis");
  YYERROR;
}

|  '(' allExpr EOI %prec '=' {
  Sdberror("unmatched left parenthesis");
  YYERROR;
}

| '(' allStrings EOI %prec '=' {
  Sdberror("unmatched left parenthesis");
  YYERROR;
}


%% /* --------------------------------------------------------- */

static SdbVariable *VarPtr=NULL;

#include "lex.Sdb.c"


SdbNode*
SdbParse(interp, dbPtr, string) 
  Tcl_Interp *interp; 
  SdbFile *dbPtr;
  char *string;
{
  int result;

  /* set the following static variables so that they may be accessed
     from within the parser. */

  VarPtr = dbPtr->varPtr;
  CurrentInterp = interp;
  htbl=dbPtr->htbl;
  expr = NULL;

  analyze_buff_init(string);
  result=Sdbparse();

  /*
   * Cheack the result returned from SdbParse.  If it is non-zero
   * there has been an error.  In this case, erase all allocated
   * memeory and return NULL.
   */
  if (result) {
    safeFreeMemory();
    safeFreeLists();
    return NULL;
  }

  safeFreeLists();
  return expr;

}
